<!doctype html>
<html lang="{{ site.lang | default: 'en-US' }}">
  <head>
    <meta charset="utf-8" />
    <meta http-equiv="X-UA-Compatible" content="IE=edge" />
    <meta name="viewport" content="width=640" />

    <meta name="author" content="{{ site.github.owner_name }}" />
    <!-- Open Graph Meta Tags -->
    <meta
      property="og:title"
      content="{{ site.title | default: site.github.repository_name }}"
    />
    <meta
      property="og:description"
      content="A recreation of the Heroes of Might and Magic II game engine. This open source multiplatform project, written from scratch, is designed to reproduce the original game with significant
    improvements in gameplay, graphics and logic (including support for high-resolution graphics, improved AI, numerous fixes
    and user interface improvements), breathing new life into one of the most addictive turn-based strategy games."
    />
    <meta
      property="og:image"
      content="{{ '/images/screenshots/screenshot_world_map.png' | absolute_url }}"
    />
    <meta property="og:url" content="{{ page.url | absolute_url }}" />
    <meta property="og:type" content="website" />
    <meta property="og:site_name" content="fheroes2" />
    <meta property="article:author" content="{{ site.github.owner_name }}" />
    <meta property="article:publisher" content="{{ site.github.owner_url }}" />
    {% if page.date %}
    <meta
      property="article:published_time"
      content="{{ page.date | date_to_xmlschema }}"
    />
    {% endif %} {% if page.last_modified_at %}
    <meta
      property="article:modified_time"
      content="{{ page.last_modified_at | date_to_xmlschema }}"
    />
    {% endif %}

    <link
      rel="stylesheet"
      href="{{ '/assets/css/style.css?v=' | append: site.github.build_revision | relative_url }}"
      media="screen"
    />

    <!-- Scripts -->
    <script src="{{ '/assets/js/accessibility-manager.js?v=' | append: site.github.build_revision | relative_url }}"></script>
    <script src="{{ '/assets/js/theme-manager.js?v=' | append: site.github.build_revision | relative_url }}"></script>
    <script src="{{ '/assets/js/templates.js?v=' | append: site.github.build_revision | relative_url }}"></script>
    <script src="{{ '/assets/js/heading-copy-manager.js?v=' | append: site.github.build_revision | relative_url }}"></script>
    <script src="{{ '/assets/js/code-copy-manager.js?v=' | append: site.github.build_revision | relative_url }}"></script>

    {% seo %}
    <link
      rel="shortcut icon"
      type="image/x-icon"
      href="{{ '/images/favicon.ico' | relative_url }}"
    />
  </head>

  <body>
    <div class="shell">
      <div class="theme-toggle-container">
        <button
          id="theme-toggle"
          class="btn btn-primary"
          aria-label="Toggle dark mode"
          aria-pressed="false"
          tabindex="0"
        >
          🌙 Dark Mode
        </button>
      </div>
      <header>
        <h1>
          <a
            href="{{ site.baseurl }}/"
            data-text="{{ site.title | default: site.github.repository_name }}"
            >{{ site.title | default: site.github.repository_name }}</a
          >
        </h1>
        <h2>
          A recreation of the <strong>Heroes of Might and Magic II</strong> game
          engine.
        </h2>
      </header>

      <section id="main_content">{{ content }}</section>

      <footer>
        {% if site.github.is_project_page %}
        <p>
          This project by
          <a href="{{ site.github.owner_url }}">{{ site.github.owner_name }}</a>
          can be found on <a href="{{ site.github.repository_url }}">GitHub</a>
        </p>
        {% endif %} {% if site.github.is_user_page %}
        <p>
          Projects by
          <a href="{{ site.github.owner_url }}">{{ site.github.owner_name }}</a>
          can be found on <a href="{{ site.github.repository_url }}">GitHub</a>
        </p>
        {% endif %}
      </footer>
    </div>
  </body>
</html>
