$light: (
  primary-color: #8b4513,
  header-color: #8b4513,
  secondary-color: #daa520,
  page-color: #fff,
  text-color: #2c1810,
  background-color: #f5deb3,
  border-color: #8b4513,
  highlight-color: #cd853f,
  code-background-color: #e6e1da,
  inverse-text-color: #e0e0e0,
  inverse-background-color: #383838,
  inverse-border-color: #c0c0c0,
  inverse-primary-color: #c0c0c0,
  border-radius: 10px,
  body-text-shadow: none,
  header-text-shadow: -1px 2px 1px #cd853f,
  table-text-color: #fff,
);
