/***************************************************************************
 *   fheroes2: https://github.com/ihhub/fheroes2                           *
 *   Copyright (C) 2019 - 2025                                             *
 *                                                                         *
 *   Free Heroes2 Engine: http://sourceforge.net/projects/fheroes2         *
 *   Copyright (C) 2009 by Andrey Afletdinov <fheroes2@gmail.com>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#pragma once

#include <bitset>
#include <cstddef>
#include <cstdint>
#include <vector>

class IStreamBase;
class OStreamBase;

inline constexpr size_t numOfPuzzleTiles{ 48 };

class Puzzle : public std::bitset<numOfPuzzleTiles>
{
public:
    Puzzle();

    Puzzle & operator=( const char * str );

    void Update( uint32_t open, uint32_t total );
    void ShowMapsDialog() const;

    std::vector<uint8_t> zone1_order{ 0, 1, 2, 3, 4, 5, 6, 11, 12, 17, 18, 23, 24, 29, 30, 35, 36, 41, 42, 43, 44, 45, 46, 47 };
    std::vector<uint8_t> zone2_order{ 7, 8, 9, 10, 13, 16, 19, 22, 25, 28, 31, 34, 37, 38, 39, 40 };
    std::vector<uint8_t> zone3_order{ 14, 15, 32, 33 };
    std::vector<uint8_t> zone4_order{ 20, 21, 26, 27 };
};

OStreamBase & operator<<( OStreamBase & stream, const Puzzle & pzl );
IStreamBase & operator>>( IStreamBase & stream, Puzzle & pzl );
