$dark: (
  primary-color: #c0c0c0,
  header-color: #e4c814,
  secondary-color: #480a0a,
  page-color: #989898,
  text-color: #e0e0e0,
  background-color: #383838,
  border-color: #480a0a,
  highlight-color: #800080,
  code-background-color: #2a2a2a,
  inverse-text-color: #2c1810,
  inverse-background-color: #f5deb3,
  inverse-border-color: #8b4513,
  inverse-primary-color: #8b4513,
  border-radius: 5px,
  body-text-shadow: 2px 2px 2px #383838,
  header-text-shadow: -2px 3px 1px darken(#a42da4, 10%),
  table-text-color: #e0e0e0,
);
